import os
import requests
from dotenv import load_dotenv
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json import json_loads

DEFAULT_MAX_TOKENS = 2048

class API_Perplexity:

    def __init__(self):
        self.url         = "https://api.perplexity.ai/chat/completions"
        self.model       = "mistral-7b-instruct"
        self.temperature = 0
        self.stream      = True                                             # todo: add support for non stream response and check the performance implications of streaming

# avaiable models and pricing (as of Dec 2024) https://docs.perplexity.ai/docs/model-cards anf https://docs.perplexity.ai/docs/pricing
#                                                       $/1M input tokens	$/1M output tokens
#   codellama-34b-instruct	16384	Chat Completion     0.35	            $1.40
#   llama-2-70b-chat	    4096	Chat Completion     $0.70	            $2.80
#   mistral-7b-instruct	    4096 	Chat Completion     $0.07	            $0.28
#   pplx-7b-chat	        8192	Chat Completion     $0.07	            $0.28
#   pplx-70b-chat	        4096	Chat Completion     $0.70	            $2.80
#   pplx-7b-online	        4096	Chat Completion     $5 (1000 requests)  $0.28
#   pplx-70b-online         4096	Chat Completion     $5 (1000 requests)  $0.28

    @cache_on_self
    def api_key(self):
        load_dotenv()
        return os.getenv('PERPLEXITY_API')

    def ask_using_messages(self, messages, model=None, temperature=None, max_tokens=None, async_mode=True):
        generator = self.create(messages, model=model, temperature=temperature, max_tokens=max_tokens)
        if async_mode:
            return generator
        full_answer = ""

        for item in generator:
            if item is not None:
                full_answer += item
        return full_answer

    def ask_using_system_prompts(self, user_prompt, images=None, system_prompts=None, histories=None, model=None,
                                 temperature=None, seed=None, max_tokens=None, async_mode=False):
        messages = []
        if system_prompts:
            for system_prompt in system_prompts:
                messages.append({"role": "system", "content": system_prompt})
        if histories:
            for item in histories:
                question = item.question
                answer   = item.answer
                messages.append({"role": "user"     , "content": question})
                messages.append({"role": "assistant", "content": answer})
        if images:
            user_prompt  = [ { "type"      : "text"      ,
                               "text"      : user_prompt }]
            for image_url in images:
                image = { "type"     : "image_url"       ,
                         "image_url": { "url": image_url }}
                user_prompt.append(image)
        messages.append({"role": "user", "content": user_prompt})

        return self.ask_using_messages(messages, model=model, temperature=temperature, max_tokens=max_tokens,  async_mode=async_mode)


    def create(self, messages,  model=None, temperature=None, max_tokens=None):

        payload = dict(model       = model       or self.model        ,
                       messages    = messages                         ,
                       temperature = temperature or self.temperature  ,
                       stream      = self.stream                      ,
                       max_tokens  = max_tokens  or DEFAULT_MAX_TOKENS)

        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": f"Bearer {self.api_key()}"
        }

        response = requests.post(self.url, headers=headers, json=payload, stream=payload.get('stream'))
        return self.parse_response(response)

    def parse_response(self, response):
        for line in response.iter_lines():
            if line:
                chunck = line.decode('utf-8')
                if chunck.startswith('data: '):
                    data = json_loads(chunck[6:])
                    for choice in data.get('choices', []):
                        content = choice.get('delta',{}).get('content')
                        yield content



from osbot_utils.base_classes.Kwargs_To_Self        import Kwargs_To_Self
from osbot_llms.models.llm_actions.LLM__Request     import LLM__Request
from osbot_llms.models.llm_actions.LLM__Response    import LLM__Response

class LLM__Action(Kwargs_To_Self):
    request : LLM__Request
    response: LLM__Response


from osbot_llms.llms.providers.LLM__Chat_Completion    import LLM__Chat_Completion
from osbot_utils.utils.Env                             import get_env

TOGETHER_AI__LLM_BASE_URL  = "https://api.together.xyz/v1/completions/"
TOGETHER_AI__DEFAULT_MODEL = "meta-llama/Llama-3-70b-chat-hf"

class LLM__Together_AI(LLM__Chat_Completion):

    def __init__(self) -> None:
        self.api_key = get_env("TOGETHER_AI_API_KEY")
        self.base_url = TOGETHER_AI__LLM_BASE_URL
        self.model    = TOGETHER_AI__DEFAULT_MODEL
        super().__init__()


from osbot_utils.base_classes.Type_Safe import Type_Safe

# todo refactor common logic into this class
class LLM__Platform_Engine(Type_Safe):
    def is_provider_available(self):
        pass

    def execute_request(self):
        pass


from dataclasses import dataclass, field
from decimal import Decimal
from typing import Optional, List

from osbot_utils.base_classes.Type_Safe import Type_Safe
from pydantic                                   import BaseModel
DEFAULT_USER_PROMPT = 'Hi'
DEFAULT_TEMPERATURE = 0.0
DEFAULT_SEED        = 42

@dataclass
class GPT_Prompt_Simple(Type_Safe):
    images        : list                = field(default_factory=list)
    chat_thread_id: Optional[str]       = field(default='')
    user_prompt   : str                 = DEFAULT_USER_PROMPT
    temperature   : float               = float(DEFAULT_TEMPERATURE)
    seed          : int                 = DEFAULT_SEED
    max_tokens    : Optional[int]       = None
    user_data     : Optional[dict]      = None
    stream        : Optional[bool]      = True


from osbot_utils.utils.Misc                            import timestamp_utc_now
from osbot_utils.utils.Status                          import status_ok, status_error
from osbot_llms.backend.s3_minio.S3_DB_Base            import S3_DB_Base
from osbot_llms.backend.s3_minio.S3__Key__Chat_Thread  import S3__Key__Chat_Thread
from osbot_llms.models.LLMs__Chat_Completion           import LLMs__Chat_Completion

S3_BUCKET_SUFFIX__CHAT_THREADS    = 'chat-threads'
CHAT__REQUEST_TYPE__USER_REQUEST  = 'user-request'
CHAT__REQUEST_TYPE__USER_RESPONSE = 'user-response'
CHAT__REQUEST_TYPE__LLM_REQUEST   = 'llm-request'
CHAT__REQUEST_TYPE__LLM_RESPONSE  = 'llm-response'

class S3_DB__Chat_Threads(S3_DB_Base):
    bucket_name__suffix   : str = S3_BUCKET_SUFFIX__CHAT_THREADS
    save_as_gz            : bool = True
    s3_key_generator      : S3__Key__Chat_Thread


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        with self.s3_key_generator as _:
            _.root_folder      = S3_BUCKET_SUFFIX__CHAT_THREADS
            _.save_as_gz       = self.save_as_gz
            _.server_name      = self.server_name
            _.use_hours        = True
            _.use_minutes      = False
            _.use_request_path = False

    def s3_key(self, **kwargs):
        s3_key = self.s3_key_generator.create__for__chat_thread(**kwargs)
        return s3_key

    # move these files to the S3_Chat_Thread class

    def save_chat_completion__user_request(self, llm_chat_completion: LLMs__Chat_Completion, request_id: str):
        if not request_id:
            return
        request_type           = CHAT__REQUEST_TYPE__USER_REQUEST
        llm_request_id         = request_id
        chat_thread_id         = llm_chat_completion.chat_thread_id
        chat_data              = dict(request_id          = request_id                       ,
                                      chat_thread_id      = chat_thread_id                   ,
                                      llm_request_id      = llm_request_id                   ,
                                      timestamp           =  timestamp_utc_now()             ,
                                      llm_chat_completion =  llm_chat_completion.json())
        s3_key                 = self.s3_key(chat_thread_id=chat_thread_id, llm_request_id=llm_request_id, request_type=request_type)
        s3_key_items           = s3_key.split('/')
        public_chat_id         = '/'.join(s3_key_items[2:6])
        public_chat_thread__id = '/'.join(s3_key_items[2:5])

        metadata               = { 'request_id'  : request_id   ,
                                   'request_type': request_type }
        if self.s3_save_data(data=chat_data, s3_key=s3_key, metadata=metadata):
            return status_ok(data={'llm_request_id'        : llm_request_id         ,
                                   'public_chat_id'        : public_chat_id         ,
                                   'public_chat_thread__id': public_chat_thread__id ,
                                   's3_key'                :s3_key                  })
        return status_error(message="s3 save data failed")

    def save_chat_completion__user_response(self, llm_chat_completion: LLMs__Chat_Completion, request_id: str):
        request_type   = CHAT__REQUEST_TYPE__USER_RESPONSE
        chat_thread_id = llm_chat_completion.chat_thread_id
        llm_request_id = request_id
        s3_key         = self.s3_key(chat_thread_id=chat_thread_id, llm_request_id=llm_request_id, request_type=request_type)
        data           = llm_chat_completion.json()
        metadata = {'request_id'    : request_id    ,
                    'request_type'  : request_type  }
        if self.s3_save_data(data=data, s3_key=s3_key, metadata=metadata):
            return status_ok(data={'llm_request_id': llm_request_id, 's3_key':s3_key})
        return status_error(message="s3 save data failed")





from decimal                                                import Decimal
from osbot_utils.base_classes.Kwargs_To_Self                import Kwargs_To_Self
from osbot_utils.helpers.Random_Guid                        import Random_Guid
from osbot_llms.llms.providers.open_router.LLM__Open_Router import LLM__Models__Open_Router, LLM__Providers

LLM_REQUEST__DEFAULT_MAX_TOKENS     = 4092
LLM_REQUEST__DEFAULT_MODEL          = LLM__Models__Open_Router.LLAMA_3_8B__FREE
LLM_REQUEST__DEFAULT_MODEL_PROVIDER = LLM__Providers.OPEN_ROUTER
LLM_REQUEST__DEFAULT_SEED           = 42
LLM_REQUEST__DEFAULT_TEMPERATURE    = Decimal('0.1')        # use '0.1' instead of 0.1, so that we don't get Decimal('0.1000000000000000055511151231257827021181583404541015625')
LLM_REQUEST__DEFAULT_USER_PROMPT    = 'Hi'

class LLM__Request(Kwargs_To_Self):
    chat_thread_id : Random_Guid
    histories      : list
    images         : list
    max_tokens     : int     = LLM_REQUEST__DEFAULT_MAX_TOKENS
    model          : str     = LLM_REQUEST__DEFAULT_MODEL
    model_provider : str     = LLM_REQUEST__DEFAULT_MODEL_PROVIDER
    seed           : int     = LLM_REQUEST__DEFAULT_SEED
    system_prompts : list
    temperature    : Decimal
    user_data      : dict
    user_prompt    : str     = LLM_REQUEST__DEFAULT_USER_PROMPT

    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.temperature = Decimal(LLM_REQUEST__DEFAULT_TEMPERATURE)        # todo, see if we can add Decimal to the classes that can safely be handled by Kwargs_To_Self on the initial values setup


from enum                                           import Enum
from os                                             import getenv
from osbot_llms.llms.providers.LLM__Chat_Completion import LLM__Chat_Completion

LLM_BASE_URL__OPEN_ROUTER = "https://openrouter.ai/api/v1/chat/completions"

class LLM__Providers(str, Enum):
    AWS_BEDROCK = "aws-bedrock"
    GROQ        = "groq"
    OPEN_AI     = "open-ai"
    OPEN_ROUTER = "open-router"

class LLM__Models__Open_Router(str, Enum):
    GEMINI_FLASH_1_5  = 'google/gemini-flash-1.5'
    LLAMA_3_8B__FREE  = "meta-llama/llama-3-8b-instruct:free"
    CAPYBARA_7B__FREE = "nousresearch/nous-capybara-7b:free"
    OPEN_AI_4o        = 'openai/gpt-4o'
    OPEN_AI_3_5       = 'openai/gpt-3.5-turbo'

class LLM__Open_Router(LLM__Chat_Completion):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.api_key  = getenv("OPEN_ROUTER_API_KEY")
        self.base_url = LLM_BASE_URL__OPEN_ROUTER
        super().__init__()




from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Misc                       import lower

class LLMs_API:

    def __init__(self):
        pass

    @cache_on_self
    def create_from_fastapi_routes(self, fast_api):

        open_api = fast_api.openapi()
        title    = open_api.get('info').get('title')
        paths    = self.extract_paths(open_api)
        tools    = self.extract_tools_from_paths(paths)
        llms_api = {'title': title, 'paths': paths, 'tools': tools}

        #pprint(open_api)


        return llms_api

    def extract_paths(self, open_api):
        paths = {}
        for path, methods in open_api.get('paths').items():
            for method, details in methods.items():
                params_query = {}
                params_path  = {}
                params_other = {}                               # todo: see if we need this
                name         = lower(details.get('summary').replace(' ', '_'))
                tool = {
                        "id"          : details.get('operationId'),
                        "description" : details.get('description'),
                        "method"      : method                    ,
                        "path"        : path                      ,
                        "params_query": params_query              ,
                        "params_path" : params_path               ,
                        "params_other": params_other              ,
                        "summary"     : details.get('summary'    ),
                    }
                for parameter in details.get('parameters', []):
                    title = parameter.get('schema').get('title')
                    param = { parameter.get('name') : {'description': title}}

                    if   parameter.get('in') == 'query': params_query.update(param)
                    elif parameter.get('in') == 'path' : params_path .update(param)
                    else                               : params_other.update(param)
                paths[name]=tool
        return paths


    def extract_tools_from_paths(self, paths):
        tools = []
        for path_name, path in paths.items():
            name        = path_name
            description = path.get('summary')  #+ ':\n' + path.get('description')
            params_other = path.get('params_other')
            params_path  = path.get('params_path')
            params_query = path.get('params_query')
            properties   = {**params_other , **params_path, **params_query}
            if properties:
                parameters = { "type" : "object"         ,
                               "properties": properties  }
            else:
                parameters = {}
            tool =  { "type"      : "function"                    ,
                      "function"  : { "name"       : name        ,
                                      "description": description ,
                                      "parameters" : parameters  }}
            tools.append(tool)
        return tools


from osbot_utils.base_classes.Kwargs_To_Self    import Kwargs_To_Self
from osbot_utils.helpers.Random_Guid            import Random_Guid


class LLM__Task(Kwargs_To_Self):
    task_id     : Random_Guid
    workflow_id : Random_Guid


from osbot_utils.base_classes.Type_Safe                 import Type_Safe
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_llms.OSBot_LLMs__Server_Config               import osbot_llms__server_config
from osbot_llms.backend.s3_minio.S3_DB__Chat_Threads    import S3_DB__Chat_Threads

class OSBot_LLMs__Shared_Objects(Type_Safe):

    @cache_on_self
    def s3_db_chat_threads(self):                                               # todo: refactor this code with the method s3_db_server_requests() since 95% is the same
        server_name = osbot_llms__server_config.server_name
        use_minio   = osbot_llms__server_config.s3_s3_use_minio                 # setup use of minio (vs aws)
        kwargs      = dict(use_minio   = use_minio   ,
                           server_name =  server_name)
        with  S3_DB__Chat_Threads(**kwargs)  as _:
            _.setup()                                                           # set up tasks, including creating target bucket if it doesn't exist
            _.s3_key_generator.use_request_path = False
            _.s3_key_generator.use_when         = True
            _.s3_key_generator.use_hours        = True
            _.s3_key_generator.use_minutes      = False
            return _

osbot_llms__shared_objects = OSBot_LLMs__Shared_Objects()



from osbot_utils.base_classes.Type_Safe             import Type_Safe
from osbot_llms.llms.providers.LLM__Execute_Request import LLM__Execute_Request

#MAX_ANSWER_SIZE = 512
MAX_ANSWER_SIZE = 65536                         # this in practice removes any size restrictions # todo: add feature to enable this only for the billing mode

class LLM__Chat_Completion(Type_Safe):
    base_url : str
    api_key  : str
    model    : str
    messages : list
    stream   : bool = True

    def add_histories(self, histories):
        if type(histories) is list:
            for item in histories:
                question = item.question
                answer   = item.answer
                if answer and len(answer) > MAX_ANSWER_SIZE:                        # todo: move this to a separate method and make this more scientific (for example using an LLM to create a summary)
                    answer = answer[:MAX_ANSWER_SIZE] + "..."               # do this so that we don't have an exponential growth in the size of the content sent to the LLMs
                self.add_message('user'     , question)
                self.add_message('assistant', answer)

    def add_message__system(self, system_prompt):
        self.add_message('system', system_prompt)
        return self

    def add_message__user(self, user_prompt, images=None):
        if images:
            user_prompt  = [ { "type"      : "text"      ,
                               "text"      : user_prompt }]
            for image_url in images:
                image = { "type"     : "image_url"       ,
                         "image_url": { "url": image_url }}
                user_prompt.append(image)

        self.add_message('user', user_prompt)
        return self

    def add_messages__system(self, system_prompts):
        if type(system_prompts) is list:
            for system_prompt in system_prompts:
                self.add_message__system(system_prompt)
        return self

    def add_messages__user(self, user_prompts):
        for user_prompt in user_prompts:
            self.add_message__user(user_prompt)
        return self

    def add_message(self, role, content):
        message = {'role': role, 'content': content}
        self.messages.append(message)
        return self

    def chat_completion(self):
        if self.stream:
            return self.chat_completion__streamed()
        else:
            return self.chat_completion__not_streamed()

    def chat_completion__not_streamed(self):
        if self.is_provider_not_available():
            return f"Provider not available"
        messages = self.messages
        headers  =  { 'Authorization': f'Bearer {self.api_key}',
                      'Content-Type' : 'application/json'      }
        json_data = { 'messages': messages   ,
                      'model'   : self.model ,
                      'stream'  : False      }
        post_data = { 'headers': headers     ,
                      'json'   : json_data    ,
                      'url'    : self.base_url}
        response = LLM__Execute_Request(post_data=post_data).execute()
        return response

    def chat_completion__streamed(self):
        if self.is_provider_not_available():
            yield f"Provider not available"
            return
        messages = self.messages
        headers  =  { 'Authorization': f'Bearer {self.api_key}',
                      'Content-Type' : 'application/json'      }
        json_data = { 'messages': messages   ,
                      'model'   : self.model ,
                      'stream'  : True       }
        post_data = { 'headers': headers     ,
                      'json'   : json_data    ,
                      'url'    : self.base_url}
        response  = LLM__Execute_Request(post_data=post_data).requests_post__streamed(**post_data)
        #response  = self.requests_post__streamed(**post_data)
        for item in response:
            yield item

    def is_provider_available(self):
        if self.api_key:
            return True
        return False

    def is_provider_not_available(self):
        return self.is_provider_available() is False

    def json_data(self):
        payload = { "messages": self.messages,
                    "model"   : self.model  }
        return payload

    # todo: see if this is needed (since this was created for the LLM__Action__Chat_Completion which also needs a rewrite)
    def post_data(self):
        url     = self.base_url
        headers = {"Content-Type": "application/json"}
        if self.api_key:
            headers['Authorization'] = f"Bearer {self.api_key}"
        json    = self.json_data()
        return dict(url=url, headers=headers, json=json)

    # todo: see if this is needed (since this was created for the LLM__Action__Chat_Completion which also needs a rewrite)
    def send_post_data(self):
        post_data = self.post_data()
        result    = LLM__Execute_Request(post_data=post_data).make_request(post_data=post_data)
        return result

    def send_user_prompt(self, user_prompt):                # todo: see if we need this
        self.add_message__user(user_prompt)
        post_data = self.post_data()
        #pprint(post_data)
        result    =  LLM__Execute_Request(post_data=post_data).make_request(post_data=post_data)
        return result

    def set_model(self, value):
        self.model = value
        return self

    def set_stream(self, value):
        self.stream = value
        return self


from os                                             import getenv
from osbot_llms.llms.providers.LLM__Chat_Completion import LLM__Chat_Completion

ENV_NAME__GROQ_API_KEY = 'GROQ_API_KEY'
LLM_BASE_URL__GROQ     = 'https://api.groq.com/openai/v1/chat/completions'

class LLM__Groq(LLM__Chat_Completion):

    def __init__(self) -> None:
        self.api_key  = getenv(ENV_NAME__GROQ_API_KEY)
        self.base_url = LLM_BASE_URL__GROQ
        super().__init__()




# todo: add support for direct calls to Google's Gemini

# def test_call_google_gemini(self):
#     api_key = getenv("GOOGLE_GEMINI__API_KEY")
#
#     # GENAI_API_DISCOVERY_URL = "https://generativelanguage.googleapis.com/$discovery/rest"
#     # uri = f"{GENAI_API_DISCOVERY_URL}?version=v1beta&key={api_key}"
#     # response = requests.get(uri)
#     # pprint(response.json())
#     # return
#     # url = "https://generativelanguage.googleapis.com/v2/models/?key=" + API_KEY
#     # models = requests.get(url)
#     # pprint(models.json())
#     # return
#     url = "https://generativelanguage.googleapis.com/v1beta/models/gemini-1.5-flash-latest:generateContent?key=" + api_key
#     data = {"contents": [{"parts": [{"text": "What is 2+2"}]}]}
#     with print_duration():
#         response = requests.post(url, json=data)
#
#     pprint(response.text)


from osbot_utils.testing.Logging                            import Logging
from starlette.responses                                    import StreamingResponse, Response
from fastapi                                                import Request, UploadFile, File
from osbot_utils.utils.Misc                                 import base64_to_bytes, bytes_to_base64
from osbot_llms.fast_api.Fast_API_Route                     import Fast_API__Routes
from osbot_llms.llms.API_Open_AI                            import API_Open_AI
from osbot_llms.llms.storage.Chats_Storage__S3_Minio        import Chats_Storage__S3_Minio
from osbot_llms.models.GPT_Audio_To_Text                    import GPT_Audio_To_Text
from osbot_llms.models.GPT_Prompt_With_System_And_History   import GPT_Prompt_With_System_And_History
from osbot_llms.models.GPT_Text_To_Audio                    import GPT_Text_To_Audio
from osbot_llms.models.LLMs__Chat_Completion                import LLMs__Chat_Completion

HEADER_NAME__CHAT_ID        = 'osbot-llms-chat-id'
HEADER_NAME__CHAT_THREAD_ID = 'osbot-llms-thread-id'

# todo: refactor this out since this is current only used by the Routes__Chat (via the chat/completions endpoint)
class Routes__OpenAI(Fast_API__Routes):
    path_prefix             : str = 'open_ai'
    chats_storage_s3_minio  : Chats_Storage__S3_Minio

    def __init__(self):
        super().__init__()
        self.api_open_ai = API_Open_AI().setup()
        self.logging     = Logging()


    async def prompt_with_system__not_stream(self, gpt_prompt_with_system_and_history: GPT_Prompt_With_System_And_History, request: Request):
        try:
            if self.api_open_ai.open_ai_not_available():
                return "OpenAI not available (need API Key)"
            if gpt_prompt_with_system_and_history.user_data:            # todo: refactor to use new mode
                model = gpt_prompt_with_system_and_history.user_data.get('selected_model', 'gpt-4o')
            else:
                model = 'gpt-4o'
            user_prompt     = gpt_prompt_with_system_and_history.user_prompt
            images          = gpt_prompt_with_system_and_history.images
            system_prompts  = gpt_prompt_with_system_and_history.system_prompts
            histories       = gpt_prompt_with_system_and_history.histories
            #model           = gpt_prompt_with_system_and_history.model.value
            temperature     = gpt_prompt_with_system_and_history.temperature
            seed            = gpt_prompt_with_system_and_history.seed
            max_tokens      = gpt_prompt_with_system_and_history.max_tokens
            async_mode      = False

            gpt_response    = await self.api_open_ai.ask_using_system_prompts(user_prompt=user_prompt,
                                                                              images=images,
                                                                              system_prompts=system_prompts,
                                                                              histories=histories,
                                                                              model=model,
                                                                              temperature=temperature,
                                                                              seed=seed,
                                                                              max_tokens=max_tokens,
                                                                              async_mode=async_mode)

            request_headers = {key: value for key, value in request.headers.items()}            # todo: refactor out
            #log_llm_chat(gpt_prompt_with_system_and_history, gpt_response, request_headers)
            return gpt_response
        except Exception as error:
            return f'Error: {error}'

    async def prompt_with_system__stream(self, gpt_prompt_with_system_and_history: GPT_Prompt_With_System_And_History, request: Request):
        request_id       = self.request_id(request)
        chat_save_result = self.chats_storage_s3_minio.save_user_request(gpt_prompt_with_system_and_history,request_id)

        async def streamer():
            if self.api_open_ai.open_ai_not_available():
                yield "OpenAI not available (need API Key)"
                return
            if gpt_prompt_with_system_and_history.user_data:            # todo: refactor to use new mode
                model = gpt_prompt_with_system_and_history.user_data.get('selected_model', 'gpt-4o')
            else:
                model = 'gpt-4o'
            user_prompt     = gpt_prompt_with_system_and_history.user_prompt
            images          = gpt_prompt_with_system_and_history.images
            system_prompts  = gpt_prompt_with_system_and_history.system_prompts
            histories       = gpt_prompt_with_system_and_history.histories
            #model           = gpt_prompt_with_system_and_history.model.value
            temperature     = gpt_prompt_with_system_and_history.temperature
            seed            = gpt_prompt_with_system_and_history.seed
            max_tokens      = gpt_prompt_with_system_and_history.max_tokens
            async_mode      = True

            generator       = await self.api_open_ai.ask_using_system_prompts(user_prompt=user_prompt,
                                                                              images=images,
                                                                              system_prompts=system_prompts,
                                                                              histories=histories,
                                                                              model=model,
                                                                              temperature=temperature,
                                                                              seed=seed,
                                                                              max_tokens=max_tokens,
                                                                              async_mode=async_mode)


            gpt_response = ''
            async for answer in generator:
                if answer:
                    gpt_response += answer
                    yield f"{answer}\n"

            llm_chat_completion = LLMs__Chat_Completion(**gpt_prompt_with_system_and_history.json())
            llm_chat_completion.llm_answer = gpt_response

            self.chats_storage_s3_minio.save_user_response(llm_chat_completion, request_id)

            request_headers = {key: value for key, value in request.headers.items()}
            #log_llm_chat(gpt_prompt_with_system_and_history, gpt_response, request_headers)


            #self.logging          .add_prompt_request(gpt_prompt_with_system_and_history, gtp_response, request_headers)  # todo: remove this one once the dydb_chat_threads is working

        response = StreamingResponse(streamer(), media_type='text/event-stream"; charset=utf-8')
        response.headers.append(HEADER_NAME__CHAT_ID       , chat_save_result.get('public_chat_id'        , ''))
        response.headers.append(HEADER_NAME__CHAT_THREAD_ID, chat_save_result.get('public_chat_thread__id', ''))
        return response

    def request_id(self, request: Request):
        if request:
            if hasattr(request.state, "request_id"):
                return request.state.request_id


    def add_routes(self):
        @self.router.post('/prompt_with_system__stream')
        async def prompt_with_system__stream(gpt_prompt_with_system_and_history: GPT_Prompt_With_System_And_History, request: Request):  # = Depends()):
            return await self.prompt_with_system__stream(gpt_prompt_with_system_and_history, request)

        @self.router.post('/audio_to_text')
        async def audio_to_text(gpt_audio_to_text: GPT_Audio_To_Text):
            audio_base_64 = gpt_audio_to_text.audio_base_64
            audio_bytes   = base64_to_bytes(audio_base_64)
            audio_text    = self.api_open_ai.audio_to_text(audio_bytes=audio_bytes)

            return Response(content=audio_text, media_type="text/plain")

        @self.router.post('/text_to_audio')
        async def text_to_audio(gpt_text_to_audio: GPT_Text_To_Audio):
            input_text  = gpt_text_to_audio.audio_text
            audio_bytes = self.api_open_ai.text_to_audio(input_text)
            audio_base_64 = bytes_to_base64(audio_bytes)
            return Response(content=audio_base_64, media_type="text/plain")



from fastapi import APIRouter

from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self


class Fast_API__Routes(Kwargs_To_Self):
    router      : APIRouter
    path_prefix : str

    def add_routes(self):
        pass

    def setup(self, app):
        prefix = f'/{self.path_prefix}'
        tags   = [self.path_prefix]
        self.add_routes()
        app.include_router(self.router, prefix=prefix, tags=tags)


from dataclasses import dataclass
from typing                              import Optional

from osbot_llms.models.GPT_Prompt_Simple import GPT_Prompt_Simple

@dataclass
class GPT_Prompt_With_System(GPT_Prompt_Simple):
    system_prompts: Optional[list[str]] = None


from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.helpers.sqlite.domains.Sqlite__DB  import Sqlite__DB
from osbot_utils.utils.Objects import pickle_load_from_bytes

from osbot_llms.llms.API_Open_AI import API_Open_AI
from osbot_llms.llms.rag.Sqlite__Table__Embeddings import Sqlite__Table__Embeddings



TABLE_NAME__RAG__RAG_DATA       = 'rag_data'
TABLE_NAME__RAG__RAG_EMBEDDINGS = 'rag_embeddings'


class RAG__DB__Sqlite(Sqlite__DB):
    api_open_ai : API_Open_AI

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # def tables_to_add(self):
    #     return { TABLE_NAME__RAG__RAG_DATA: TABLE_SCHEMA__RAG_DATA }

    @cache_on_self
    def table_embeddings(self):
        return Sqlite__Table__Embeddings(database=self)

    def add_content(self, title, content):
        with self.table_embeddings() as _:
            if _.embedding_not_exists(title):
                return _.add_embeddings(title, content)

    def embeddings__for_key(self, key, add_if_not_found=True):
        row = self.table_embeddings().where_one(key=key)
        if row:
            embeddings_bytes = row.get('embeddings')
            return pickle_load_from_bytes(embeddings_bytes)
        else:
            if add_if_not_found:
                return self.add_content(key,key)
            else:
                return []

    def embeddings__all_rows(self):
        all_embeddings = {}
        for row in self.table_embeddings().rows():
            embeddings_bytes    = row.get('embeddings')
            key                 = row.get('key')
            embeddings          = pickle_load_from_bytes(embeddings_bytes)
            all_embeddings[key] = embeddings
        return all_embeddings

    def find_similarities__from_embeddings(self, embeddings_to_match, threshold=0.4):
        import numpy as np
        def cosine_similarity(a, b):
            return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))

        similarities = []
        for key, embeddings in self.embeddings__all_rows().items():
            similarity = cosine_similarity(embeddings_to_match, embeddings)
            if similarity > threshold:
                similarities.append((similarity, key))
        similarities.sort(reverse=True, key=lambda x: x[0])
        return similarities

    def similarities_content(self, similarities):
        result = {}
        with self.table_embeddings() as _:
            for key, similarity in similarities:
                content = _.where_one(key=key)

    def setup(self):
        super().setup()
        self.table_embeddings().setup()
        return self


from unittest import TestCase

import httpx
from fastapi                                import FastAPI, Request
from osbot_utils.helpers.Random_Guid        import Random_Guid
from osbot_utils.utils.Http                 import url_join_safe
from osbot_fast_api.utils.Fast_API_Server   import Fast_API_Server
from starlette.responses                    import StreamingResponse


class test__chained_fast_api(TestCase):

    def test_check__fast_api_2(self):

        # FastAPI_2 (Backend Service):
        app_2 = FastAPI()

        async def generate_stream_data(api_key):
            for i in range(3):
                yield f"Data chunk {i}: {api_key}\n"

        @app_2.post("/chat/completions")
        async def chat_completions(request: Request):
            request_body = await request.json()
            api_key      = request_body.get("api_key", '(no key)')
            return StreamingResponse(generate_stream_data(api_key), media_type="text/plain")


        # FastAPI_1 (Intermediate Service)
        app_1 = FastAPI()

        async def proxy_stream(url: str, json_payload: dict, headers: dict):
            async with httpx.AsyncClient() as client:
                async with client.stream("POST", url, json=json_payload, headers=headers) as response:
                    async for chunk in response.aiter_text():
                        yield chunk


        @app_1.post("/chat/completions")
        async def chat_completions(request: Request):
            request_body = await request.json()
            api_key      = f"YOUR_API_KEY : {Random_Guid()}"
            backend_url  = request_body.get("backend_url")
            request_body["api_key"] = api_key

            response  = StreamingResponse(proxy_stream(backend_url, request_body, headers={"accept": "text/plain"}),media_type="text/plain")
            response.headers['api_key'] = api_key
            return response

        with Fast_API_Server(app=app_2) as fast_api_2:
            response_1 = fast_api_2.requests_post("/chat/completions", data={})
            assert response_1.status_code == 200
            assert response_1.text        == 'Data chunk 0: (no key)\nData chunk 1: (no key)\nData chunk 2: (no key)\n'

            post_data  = dict(user_prompt = '40_2',
                              backend_url = url_join_safe(fast_api_2.url(), '/chat/completions'))
            with Fast_API_Server(app=app_1) as fast_api_1:
                response_2 = fast_api_1.requests_post("/chat/completions", data=post_data)
                assert response_2.status_code == 200
                api_key = response_2.headers.get('api_key')
                assert 'YOUR_API_KEY' in api_key
                assert response_2.text == (f'Data chunk 0: {api_key}\n'
                                           f'Data chunk 1: {api_key}\n'
                                           f'Data chunk 2: {api_key}\n')



import asyncio
import traceback

from fastapi                                                                import Request
from fastapi.params import Header, Body
from osbot_utils.utils.Dev import pprint
from starlette.responses                                                    import StreamingResponse
from osbot_fast_api.api.Fast_API_Routes                                     import Fast_API_Routes
from osbot_utils.context_managers.capture_duration                          import capture_duration
from osbot_llms.OSBot_LLMs__Shared_Objects                                  import osbot_llms__shared_objects
from osbot_llms.fast_api.routes.Routes__OpenAI                              import Routes__OpenAI
from osbot_llms.llms.chats.LLM__Chat_Completion__Resolve_Engine             import LLM__Chat_Completion__Resolve_Engine
from osbot_llms.llms.storage.Chats_Storage__S3_Minio                        import Chats_Storage__S3_Minio
from osbot_llms.models.LLMs__Chat_Completion import LLMs__Chat_Completion, SWAGGER_EXAMPLE__LLMs__Chat_Completion

ROUTES_PATHS__CONFIG        = ['/config/status', '/config/version']
HEADER_NAME__CHAT_ID        = 'osbot-llms-chat-id'
HEADER_NAME__CHAT_THREAD_ID = 'osbot-llms-thread-id'
HEADER_NAME__CHAT_PLATFORM  = 'osbot-llms-platform'
HEADER_NAME__CHAT_PROVIDER  = 'osbot-llms-provider'
HEADER_NAME__CHAT_MODEL     = 'osbot-llms-model'

class Routes__Chat(Fast_API_Routes):
    tag                     : str = 'chat'
    chats_storage_s3_minio  : Chats_Storage__S3_Minio

    def execute_llm_request(self, llm_chat_completion):
        llm_platform_engine = LLM__Chat_Completion__Resolve_Engine().map_provider(llm_chat_completion)
        if llm_platform_engine:
            return llm_platform_engine.execute_request()
        return 'no engine'

    async def handle_other_llms(self, llm_chat_completion: LLMs__Chat_Completion, request: Request, request_id: str):
        stream = llm_chat_completion.stream
        if stream:
            return StreamingResponse(self.handle_other_llms__streamer(llm_chat_completion, request, request_id), media_type='text/event-stream"; charset=utf-8')
        else:
            return await self.handle_other_llms__no_stream(llm_chat_completion, request, request_id)

    async def handle_other_llms__no_stream(self, llm_chat_completion: LLMs__Chat_Completion, request: Request, request_id: str):
        complete_answer =  self.execute_llm_request(llm_chat_completion)
        try:
            #request_headers = {key: value for key, value in request.headers.items()}
            llm_chat_completion.llm_answer = complete_answer
        except:
            pass
        return complete_answer

    async def handle_other_llms__streamer(self, llm_chat_completion: LLMs__Chat_Completion, request: Request,  request_id: str):
        with capture_duration() as duration:
            complete_answer = ''
            async def simulated_api_call():                         # Simulating the response of the async API call
                #user_data = llm_chat_completion.user_data or {}

                response =  self.execute_llm_request(llm_chat_completion)
                for chunk in response:
                    await asyncio.sleep(0)                           # this is needed to trigger sending the data back (without it, we don't get streaming)
                    yield chunk

            generator = simulated_api_call()
            async for answer in generator:
                if answer:
                    complete_answer += answer
                    yield f"{answer}\n"

        try:
            llm_chat_completion.llm_answer = complete_answer
            self.chats_storage_s3_minio.save_user_response(llm_chat_completion, request_id)
            # if hasattr(request.state, "http_events"):                                 # todo: see if this is still needed (originaly this handled the case when the trace_calls did not handled ok the streamed responses
            #     http_events : Fast_API__Http_Events = request.state.http_events
            #     http_events.on_response_stream_completed(request)             # todo: rewrite this
            #print(f">>>>>>>>>>>>>>> STREAM COMPLETED: {request.state._state}")
        except Exception as error:
            print(f"[error][handle_other_llms__streamer] : {error}")        # todo: log this error
            traceback.print_exc()


    async def completion(self, request: Request, llm_chat_completion: LLMs__Chat_Completion = SWAGGER_EXAMPLE__LLMs__Chat_Completion):
        request_id       = self.request_id(request)
        chat_save_result = self.chats_storage_s3_minio.save_user_request(llm_chat_completion, request_id)

        routes_open_ai   = Routes__OpenAI()
        user_data        = llm_chat_completion.user_data
        if user_data is None:
            user_data = dict(selected_platform = llm_chat_completion.llm_platform ,
                             selected_provider = llm_chat_completion.llm_provider ,
                             selected_model    = llm_chat_completion.llm_model    )
            llm_chat_completion.user_data = user_data

        # for now use the code in routes_open_ai.prompt_with_system__stream which is already working for OpenAI
        if user_data and 'selected_platform' in user_data and user_data.get('selected_platform') != 'OpenAI (Paid)':
            response = await self.handle_other_llms(llm_chat_completion, request, request_id)
            if type(response) == StreamingResponse:
                pass
                response.headers.append(HEADER_NAME__CHAT_ID       , chat_save_result.get('public_chat_id'        ,''))
                response.headers.append(HEADER_NAME__CHAT_THREAD_ID, chat_save_result.get('public_chat_thread__id',''))
                response.headers.append(HEADER_NAME__CHAT_PLATFORM , user_data.get('selected_platform'            ,''))
                response.headers.append(HEADER_NAME__CHAT_PROVIDER , user_data.get('selected_provider'            ,''))
                response.headers.append(HEADER_NAME__CHAT_MODEL    , user_data.get('selected_model'               ,''))
            return response
        else:
            stream = llm_chat_completion.stream
            if stream:
                return await routes_open_ai.prompt_with_system__stream(llm_chat_completion, request)
            else:
                return await routes_open_ai.prompt_with_system__not_stream(llm_chat_completion, request)

    def request_id(self, request: Request):
        if request:
            if hasattr(request.state, "request_id"):
                return request.state.request_id

    # todo: refactor this method into a new Routes class (one focused on viewing chat threads)
    async def view(self, chat_id: str):
        s3_db_chat_threads = osbot_llms__shared_objects.s3_db_chat_threads()
        server_name = s3_db_chat_threads.server_name
        s3_key = f'chat-threads/{server_name}/{chat_id}/user-response.json.gz'
        return osbot_llms__shared_objects.s3_db_chat_threads().s3_file_data(s3_key)

    def setup_routes(self):
        self.router.post("/completion")(self.completion )
        self.router.get("/view"       )(self.view)



import requests
from enum                                           import Enum
from urllib.parse                                   import urljoin
from osbot_llms.llms.providers.LLM__Chat_Completion import LLM__Chat_Completion

ENV_KEY__OLLAMA__SERVER = 'OLLAMA__SERVER'                  # todo: add support for defining the IP and port and schema of the ollama server
LLM_BASE_URL__OLLAMA = "http://localhost:11434"

class LLM__Models__OLLAMA(str, Enum):
    LLAMA3    = 'llama3'
    PHI3      = "phi3"
    GEMMA_2B  = "gemma"



class LLM__Ollama(LLM__Chat_Completion):
    model    : str = LLM__Models__OLLAMA.GEMMA_2B
    base_url : str =  LLM_BASE_URL__OLLAMA

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def api_generate(self, prompt, model="llama3", stream=False, format=None):
        path      = 'api/generate'
        json_data = { 'model'  :  model,
                      'prompt' : prompt   ,
                      'stream' : stream   ,
                      "options": {
                          "seed"          : 42,
                          "temperature"   : 0 }}
        if format:
            json_data['format'] = format
        return self.post_request(path, json_data, stream=stream)

    def model_details(self, model_name):
        return self.post_request('/api/show', { 'name' : model_name })

    def models(self):
        return self.get_request('/api/tags').get('models')

    # def post_request(self, path, json_data):
    #     url       = urljoin(self.base_url, path)
    #     headers   = {"Content-Type": "application/json"}
    #     post_data = dict(url=url, headers=headers, json=json_data)
    #     return self.make_request(post_data)

    def post_request(self, path, json_data, stream=False):
        url = urljoin(self.base_url, path)
        headers = {"Content-Type": "application/json"}
        with requests.post(url, headers=headers, json=json_data, stream=stream) as response:
            if stream:
                for line in response.iter_lines():
                    if line:
                        yield line.decode('utf-8')
            else:
                return response.json()

    def get_request(self, path='/'):
        url = urljoin(self.base_url, path)
        response = requests.get(url)
        if response.headers['Content-Type'] == 'text/plain; charset=utf-8':
            return response.text
        return response.json()

    def post_data(self):
        post_data = super().post_data()
        post_data['url'] += '/api/chat'
        post_data['json']['stream'] = False
        return post_data


from dataclasses import dataclass
from typing                                                import Optional

from osbot_llms.models.GPT_History            import GPT_History
from osbot_llms.models.GPT_Prompt_With_System import GPT_Prompt_With_System

@dataclass
class GPT_Prompt_With_System_And_History(GPT_Prompt_With_System):
    histories      : Optional[list[GPT_History]] = None


from os import getenv
from osbot_llms.llms.providers.LLM__Chat_Completion import LLM__Chat_Completion

LLM_PLATFORM_NAME           = 'SambaNova (Free)'
ENV_NAME__SAMBANOVA_API_KEY = 'SAMBANOVA_API_KEY'
LLM_BASE_URL__MISTRAL       = 'https://api.sambanova.ai/v1/chat/completions'
SAMBANOVA__DEFAULT_MODEL    = 'Meta-Llama-3.1-405B-Instruct'

class LLM__SambaNova(LLM__Chat_Completion):

    def __init__(self) -> None:
        self.api_key  = getenv(ENV_NAME__SAMBANOVA_API_KEY)
        self.base_url = LLM_BASE_URL__MISTRAL
        self.model    = SAMBANOVA__DEFAULT_MODEL
        super().__init__()




from fastapi                                        import Request, Response
from osbot_utils.base_classes.Type_Safe             import Type_Safe
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self

from osbot_llms.OSBot_LLMs__Server_Config  import osbot_llms__server_config
from osbot_llms.OSBot_LLMs__Shared_Objects import osbot_llms__shared_objects


class Chats_Storage__S3_Minio(Type_Safe):

    def __init__(self):
        super().__init__()

    @cache_on_self
    def s3_log_requests(self):
        return osbot_llms__server_config.s3_log_requests

    @cache_on_self
    def s3_db_chat_threads(self):
        return osbot_llms__shared_objects.s3_db_chat_threads()

    def save_user_request(self, llm_chat_completion, request_id):
        if osbot_llms__server_config.s3_log_requests:
            result = osbot_llms__shared_objects.s3_db_chat_threads().save_chat_completion__user_request(llm_chat_completion, request_id)
            if result.get('status') == 'ok':
                return result.get('data')
        return {}

    def save_user_response(self, llm_chat_completion, request_id):
        if self.s3_log_requests():
            return self.s3_db_chat_threads().save_chat_completion__user_response(llm_chat_completion,request_id)
        return {}


from osbot_utils.utils.Status                           import status_warning
from osbot_llms.backend.s3_minio.S3_DB_Base__Disabled   import S3_DB_Base__Disabled
from osbot_llms.backend.s3_minio.S3_DB__Chat_Threads    import S3_DB__Chat_Threads
from osbot_llms.models.LLMs__Chat_Completion            import LLMs__Chat_Completion

class S3_DB__Chat_Threads__Disabled(S3_DB__Chat_Threads, S3_DB_Base__Disabled):

    def s3_key(self, **kwargs):
        return None

    def save_chat_completion__user_request(self, llm_chat_completion: LLMs__Chat_Completion, request_id: str):
        return status_warning("Warning: using S3_DB__Chat_Threads__Disabled, so no data will be saved.")

    def save_chat_completion__user_response(self, llm_chat_completion: LLMs__Chat_Completion, request_id: str):
        return status_warning("Warning: using S3_DB__Chat_Threads__Disabled, so no data will be saved.")


from osbot_utils.utils.Misc                         import random_guid
from osbot_llms.backend.s3_minio.S3__Key_Generator  import S3__Key_Generator

VALUE__REQUEST_TYPE__UNKNOWN = 'unknown'

class S3__Key__Chat_Thread(S3__Key_Generator):
    use_request_path  : bool = False

    def create__for__chat_thread(self, chat_thread_id=None, llm_request_id=None, request_type=None):
        path_elements = self.create_path_elements__from_when()

        if not chat_thread_id:
            chat_thread_id = random_guid()
        if not llm_request_id:
            llm_request_id = random_guid()
        if not request_type:
            request_type = VALUE__REQUEST_TYPE__UNKNOWN
        path_elements.append(chat_thread_id)
        path_elements.append(llm_request_id)

        file_id = request_type

        s3_key = self.create_s3_key(path_elements=path_elements, file_id=file_id)
        return s3_key




import requests
from osbot_utils.utils.Env                          import get_env
from osbot_utils.utils.Json                         import from_json_str
from osbot_llms.llms.chats.LLM__Platform_Engine     import LLM__Platform_Engine
from osbot_llms.llms.providers.ollama.LLM__Ollama   import LLM__Ollama
from osbot_llms.models.LLMs__Chat_Completion        import LLMs__Chat_Completion

OLLAMA__BASE_URL          = "http://localhost:11434"
ENV_VAR__OLLAMA__BASE_URL = 'OLLAMA__BASE_URL'

class LLM__Platform_Engine__Ollama(LLM__Platform_Engine):
    llm_platform       : str
    llm_provider       : str
    llm_model          : str
    llm_chat_completion: LLMs__Chat_Completion

    def is_provider_available(self):
        try:
            response = requests.get(self.ollama_base_url())
            if response.text == "Ollama is running":
                return True
        except requests.RequestException:
            pass
        return False

    def llm_ollama(self):
        return LLM__Ollama(base_url=self.ollama_base_url())

    def ollama_base_url(self):
        return get_env(ENV_VAR__OLLAMA__BASE_URL)  or OLLAMA__BASE_URL

    # todo: add support for non streamed requests
    def execute_request(self):
        if self.is_provider_available():
            try:
                request_kwargs = dict(prompt = self.llm_chat_completion.user_prompt,
                                      model  = self.llm_model                      ,
                                      stream = True                                )
                result = self.llm_ollama().api_generate(**request_kwargs)
                for json_line in result:
                    if json_line:
                        chunk = from_json_str(json_line)
                        yield chunk.get('response')
            except Exception as error:
                yield f"Error when execution Ollama request: {error}"
        else:
            yield "Ollama is not available"


from osbot_utils.utils.Env                          import get_env
from osbot_llms.llms.providers.LLM__Chat_Completion import LLM__Chat_Completion

LLM_PLATFORM_NAME         = 'Mistral (Free)'
ENV_NAME__MISTRAL_API_KEY = 'MISTRAL_API_KEY'
LLM_BASE_URL__MISTRAL     = 'https://api.mistral.ai/v1/chat/completions'
MISTRAL__DEFAULT_MODEL    = 'pixtral-12b-2409'

class LLM__Mistral(LLM__Chat_Completion):

    def __init__(self) -> None:
        self.api_key  = get_env(ENV_NAME__MISTRAL_API_KEY)
        self.base_url = LLM_BASE_URL__MISTRAL
        self.model    = MISTRAL__DEFAULT_MODEL
        super().__init__()




from unittest                                            import TestCase
from osbot_aws.testing.Temp__Random__AWS_Credentials     import Temp__Random__AWS_Credentials
from osbot_utils.helpers.Random_Guid                     import Random_Guid
from osbot_llms.OSBot_LLMs__Server_Config                import osbot_llms__server_config
from osbot_llms.OSBot_LLMs__Shared_Objects               import osbot_llms__shared_objects
from osbot_llms.backend.s3_minio.S3_DB_Base              import ENV_NAME__USE_MINIO_AS_S3
from osbot_llms.models.LLMs__Chat_Completion             import LLMs__Chat_Completion


class TestCase__S3_Minio__Temp_Chat_Threads(TestCase):

    @classmethod
    def setUpClass(cls):
        osbot_llms__server_config.s3_log_requests = True       # set value of s3_log_requests

        cls.extra_env_vars        = {ENV_NAME__USE_MINIO_AS_S3: 'True'}
        cls.random_aws_creds      = Temp__Random__AWS_Credentials(env_vars=cls.extra_env_vars).set_vars()
        cls.s3_db_chat_threads    = osbot_llms__shared_objects.s3_db_chat_threads   (reload_cache=True)
        cls.server_name           = osbot_llms__server_config.server_name

        with cls.s3_db_chat_threads as _:                       # setup bucket for chat_threads
            assert _.using_minio() is True                      # confirm we are using Minio
            assert _.setup      () is _                         # this will create the temp bucket
            assert _.bucket_exists() is True

    @classmethod
    def tearDownClass(cls):
        with cls.s3_db_chat_threads as _:                       # delete temp bucket for chat_threads
            assert _.using_minio() is True
            assert _.bucket_delete_all_files()
            assert _.bucket_delete() is True

        cls.random_aws_creds.restore_vars()

        osbot_llms__server_config.s3_log_requests = False      # restore value of s3_log_requests

    def s3_key__new_chat(self, **kwargs):
        request_id = Random_Guid()
        llm_chat_completion = LLMs__Chat_Completion(**kwargs)
        response = self.s3_db_chat_threads.save_chat_completion__user_response(llm_chat_completion, request_id)
        return response.get('data').get('s3_key')

    def s3_key__new_chats(self, count=5, **kwargs):
        s3_keys = []
        for i in range(count):
            s3_keys.append(self.s3_key__new_chat(**kwargs))
        return s3_keys


from osbot_utils.base_classes.Type_Safe         import Type_Safe
from osbot_utils.helpers.sqlite.Sqlite__Table   import Sqlite__Table
from osbot_utils.utils.Misc                     import timestamp_utc_now
from osbot_utils.utils.Objects                  import pickle_save_to_bytes, pickle_load_from_bytes
from osbot_llms.llms.API_Open_AI                import API_Open_AI

SQLITE__TABLE_NAME__EMBEDDINGS  = 'embeddings'       # note: table name doesn't support the - char
EMBEDDINGS__DEFAULT__DIMENSIONS = 256

class Schema__Table__Embeddings(Type_Safe):
    key         : str           # index key
    content     : str
    dimensions  : int
    embeddings  : bytes          # todo: for now using pickle (see performance implication) and figure out a way to store the embeddings in a more native way (when trying to convert the embeddings int values into bytes and back to ints, the values didn't seem to be matching)
    model       : str
    provider    : str
    platform    : str
    timestamp   : int
    token_count : int


class Sqlite__Table__Embeddings(Sqlite__Table):
    api_open_api : API_Open_AI

    def __init__(self, **kwargs):
        self.table_name = SQLITE__TABLE_NAME__EMBEDDINGS
        self.row_schema  = Schema__Table__Embeddings
        super().__init__(**kwargs)

    def setup(self):
        if self.exists() is False:
            self.create()
            self.index_create('key')
        return self

    def embedding(self, key):
        return self.where_one(key=key)

    def embedding_not_exists(self, key):
        return self.embedding(key) is None

    def add_embeddings(self, key, content):
        platform         = "OpenAI (Paid)"
        provider         = "OpenAI"
        model            = 'text-embedding-3-small'
        dimensions       = EMBEDDINGS__DEFAULT__DIMENSIONS
        result           = self.api_open_api.embeddings(content, model, dimensions=dimensions)
        embeddings_data  = result.get('embeddings_data')
        embeddings       = pickle_save_to_bytes(embeddings_data)        # todo: document the security implications of using pickle
        token_count      = result.get('total_tokens'  )

        row_data = dict(key        = key                ,
                        content    = content            ,
                        dimensions = dimensions         ,
                        embeddings = embeddings         ,
                        model      = model              ,
                        provider   = provider           ,
                        platform   = platform           ,
                        timestamp  = timestamp_utc_now(),
                        token_count= token_count        )
        self.add_row_and_commit(**row_data)
        return embeddings_data



from osbot_llms.llms.chats.LLM__Platform_Engine                 import LLM__Platform_Engine
from osbot_llms.llms.providers.together_ai.LLM__Together_AI     import LLM__Together_AI
from osbot_llms.models.LLMs__Chat_Completion                    import LLMs__Chat_Completion


class LLM__Platform_Engine__Together_AI(LLM__Platform_Engine):
    llm_platform       : str
    llm_provider       : str
    llm_model          : str
    llm_chat_completion: LLMs__Chat_Completion
    llm_together_ai    : LLM__Together_AI

    def execute_request(self):
        with self.llm_together_ai as _:
            _.add_messages__system(self.llm_chat_completion.system_prompts)
            _.add_histories       (self.llm_chat_completion.histories     )
            _.add_message__user   (self.llm_chat_completion.user_prompt   )
            _.set_model           (self.llm_model                         )
            _.set_stream          (self.llm_chat_completion.stream        )
            return _.chat_completion()






from osbot_utils.utils.Dev import pprint

from osbot_llms.llms.actions.LLM__Action__Chat_Completion import LLM__Action__Chat_Completion
from osbot_llms.llms.providers.open_router.LLM__Open_Router import LLM__Open_Router
from osbot_llms.llms.tasks.LLM__Task import LLM__Task
from osbot_llms.models.llm_actions.LLM__Request import LLM__Request

# todo: needs refactoring to new @flow/@task driven approach
class LLM__Task__Chat__Completion(LLM__Task):

    def __init__( self, **kwargs ):
        super().__init__(**kwargs)

    def execute(self, llm_request: LLM__Request):
        llm_chat_completion = LLM__Open_Router()
        action_kwargs       = dict(llm_chat_completion = llm_chat_completion,
                                   llm_request         = llm_request        ,
                                   task_id             = self.task_id       ,
                                   workflow_id         = self.workflow_id   )
        action              = LLM__Action__Chat_Completion(**action_kwargs)
        action_result       = action.execute()
        pprint(action_result)
        action_status       = action_result.get('action_status' )
        action_outcome      = action_result.get('action_outcome')

        task_result = dict(status  = action_status,
                           **action_outcome)
        return task_result




from dataclasses import dataclass
from typing import Optional

from fastapi.params import Body

from osbot_llms.models.GPT_Prompt_With_System_And_History import GPT_Prompt_With_System_And_History

SWAGGER_EXAMPLE__LLMs__Chat_Completion  = Body(..., example=dict(user_prompt    ='Good morning, what is 44-2?' ,
                                                                        system_prompts = ['use emojis in the answer' ],
                                                                        #temperature    = 0.0                         ,
                                                                        seed            = 42                          ,
                                                                        stream          = False                       ,
                                                                        llm_platform    = "Groq (Free)"               ,
                                                                        llm_provider    = "1. Meta"                   ,
                                                                        llm_model       = "llama-3.1-70b-versatile"   ))

@dataclass
class LLMs__Chat_Completion(GPT_Prompt_With_System_And_History):
    llm_platform: Optional[str] = None
    llm_provider: Optional[str] = None
    llm_model   : Optional[str] = None
    llm_answer  : Optional[str] = None


from osbot_llms.llms.chats.LLM__Platform_Engine     import LLM__Platform_Engine
from osbot_llms.llms.providers.groq.LLM__Groq       import LLM__Groq
from osbot_llms.models.LLMs__Chat_Completion        import LLMs__Chat_Completion


class LLM__Platform_Engine__Groq(LLM__Platform_Engine):
    llm_platform       : str
    llm_provider       : str
    llm_model          : str
    llm_chat_completion: LLMs__Chat_Completion
    llm__groq          : LLM__Groq

    def execute_request(self):
        user_prompt = self.llm_chat_completion.user_prompt
        #images      = self.llm_chat_completion.images              # Groq doesn't seem to support images (in the same way as Open Router)
        with self.llm__groq as _:
            _.add_messages__system(self.llm_chat_completion.system_prompts)
            _.add_histories       (self.llm_chat_completion.histories     )
            _.add_message__user   (user_prompt                            )
            _.set_model           (self.llm_model                         )
            _.set_stream          (self.llm_chat_completion.stream        )
            return _.chat_completion()


import requests

from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Json import from_json_str
from osbot_utils.utils.Status import status_error


class LLM__Execute_Request(Type_Safe):
    post_data : dict


    def execute(self):
        return self.requests_post__not_streamed(**self.post_data)

    def make_request(self, post_data):
        response  = requests.post(**post_data)
        if response.status_code == 200:
            if response.headers['Content-Type'] == 'application/x-ndjson':
                return response
            else:
                return response.json()
        else:
            return status_error(error=response.text, message=f'request failed with status code: {response.status_code}')

    def requests_post__not_streamed(self, url, json, headers=None):  # todo refactor the other methods to use this naming convention
        try:
            response = requests.post(url, headers=headers, json=json)
            if response.status_code != 200:
                return  f"Error fetching provider data, status_code was {response.status_code}, text was : {response.text}"
            json_data = response.json()
            choice  = json_data.get('choices')[0]
            content = choice.get('message').get('content')
            return content
        except Exception as error:
            return f"Error fetching provider data : {error}"

    def requests_post__streamed(self, url, json, headers=None):  # todo refactor the other methods to use this naming convention
        try:
            response = requests.post(url, headers=headers, json=json, stream=True)
            for line in response.iter_lines():
                if line:
                    raw_data = line.decode('utf-8')
                    if raw_data.startswith('data: {'):
                        json_line = raw_data[5:]
                        json_data = from_json_str(json_line)
                        choice = json_data.get('choices')[0]
                        yield choice.get('delta').get('content')
                    elif raw_data.startswith('{'):
                        yield raw_data
                        # json_data = from_json_str(raw_data)
                        # if json_data:
                        #     yield json_data
        except Exception as error:
            yield f"Error fetching Open Router data : {error}"


from mangum                                 import Mangum
from osbot_llms.fast_api.Fast_API__LLMs     import Fast_API__LLMs

fast_api_llms = Fast_API__LLMs().setup()
app           = fast_api_llms.app()
run           = Mangum(app)

if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8080)




from osbot_utils.base_classes.Type_Safe                        import Type_Safe
from osbot_utils.decorators.methods.cache_on_self              import cache_on_self
from osbot_llms.llms.chats.LLM__Platform_Engine__Groq          import LLM__Platform_Engine__Groq
from osbot_llms.llms.chats.LLM__Platform_Engine__Mistral       import LLM__Platform_Engine__Mistral
from osbot_llms.llms.chats.LLM__Platform_Engine__Ollama        import LLM__Platform_Engine__Ollama
from osbot_llms.llms.chats.LLM__Platform_Engine__Open_Router   import LLM__Platform_Engine__Open_Router
from osbot_llms.llms.chats.LLM__Platform_Engine__SambaNova     import LLM__Platform_Engine__SambaNova
from osbot_llms.llms.chats.LLM__Platform_Engine__Together_AI   import LLM__Platform_Engine__Together_AI
from osbot_llms.models.LLMs__Chat_Completion                   import LLMs__Chat_Completion

USER_DATA__VAR_NAME__PLATFORM = 'selected_platform'
USER_DATA__VAR_NAME__PROVIDER = 'selected_provider'
USER_DATA__VAR_NAME__MODEL    = 'selected_model'

class LLM__Chat_Completion__Resolve_Engine(Type_Safe):

    @cache_on_self
    def llms_platform_engines(self):
        return {'Groq (Free)'        : LLM__Platform_Engine__Groq        ,
                'Ollama (Local)'     : LLM__Platform_Engine__Ollama      ,
                'Together AI (Paid)' : LLM__Platform_Engine__Together_AI ,
                'Open Router (Free)' : LLM__Platform_Engine__Open_Router ,
                'Open Router (Paid)' : LLM__Platform_Engine__Open_Router ,
                'Mistral (Free)'     : LLM__Platform_Engine__Mistral     ,
                'SambaNova (Free)'   : LLM__Platform_Engine__SambaNova   }

    def map_provider(self, llm_chat_completion: LLMs__Chat_Completion):
        llm_platform              = self.resolve__llm_platform(llm_chat_completion)       # todo: refactor out when there is no need to keep the main chat JS UI compatible with the previous API endpoint
        llm_provider              = self.resolve__llm_provider(llm_chat_completion)
        llm_model                 = self.resolve__llm_model   (llm_chat_completion)
        llm_platform_engine_class = self.llms_platform_engines().get(llm_platform)
        if llm_platform_engine_class:
            platform_kwargs     = dict(llm_platform=llm_platform, llm_provider=llm_provider,llm_model=llm_model,  llm_chat_completion=llm_chat_completion)
            llm_platform_engine = llm_platform_engine_class(**platform_kwargs)
            return llm_platform_engine
        return None
        #return f"Going to handle the request with {llm_platform} | {llm_provider} | {llm_model}"

    def resolve__llm_platform(self, llm_chat_completion: LLMs__Chat_Completion):
        user_data = llm_chat_completion.user_data or {}
        return user_data.get(USER_DATA__VAR_NAME__PLATFORM) or llm_chat_completion.llm_platform

    def resolve__llm_provider(self, llm_chat_completion: LLMs__Chat_Completion):
        user_data = llm_chat_completion.user_data or {}
        return user_data.get(USER_DATA__VAR_NAME__PROVIDER) or llm_chat_completion.llm_provider

    def resolve__llm_model(self, llm_chat_completion: LLMs__Chat_Completion):
        user_data = llm_chat_completion.user_data or {}
        return user_data.get(USER_DATA__VAR_NAME__MODEL) or llm_chat_completion.llm_model



from osbot_utils.base_classes.Type_Safe import Type_Safe


class OSBot_LLMs__Server_Config(Type_Safe):         # todo: add workflow to load this from a json/toml config file
    s3_log_requests : bool  = False
    server_name     : str   = 'osbot-llms'
    s3_s3_use_minio : bool  = True

osbot_llms__server_config = OSBot_LLMs__Server_Config()


from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_fast_api.api.Fast_API_Routes import Fast_API_Routes

from osbot_llms.utils.Version           import version__osbot_llms
from dataclasses                        import dataclass

class Routes__Info(Fast_API_Routes):
    tag : str = 'info'

    def version(self):
        return {"version" : version__osbot_llms }

    def setup_routes(self):
        self.add_route_get(self.version)


from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.helpers.Random_Guid import Random_Guid
from osbot_utils.utils.Misc                                 import timestamp_utc_now
from osbot_utils.utils.Status                               import status_ok, status_error
from osbot_llms.llms.providers.LLM__Chat_Completion         import LLM__Chat_Completion
from osbot_llms.models.llm_actions.LLM__Request             import LLM__Request


class LLM__Action__Chat_Completion(Type_Safe):
    action_id           : Random_Guid
    action_name         : str                 = 'LLM Action - Call Completion'
    action_outcome      : dict
    llm_chat_completion : LLM__Chat_Completion = None
    llm_request         : LLM__Request          = None
    llm_response        : dict
    system_prompts      : list
    action_steps        : list
    task_id             : Random_Guid
    workflow_id         : Random_Guid

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.action_steps = [self.step_1__check_llm_request     ,
                             self.step_2__create_llm_json_data  ,
                             self.step_3__llm_chat_completion   ,
                             self.step_4__extract_content       ]

    def execute(self):
        steps_result            = []
        step_data               = {}
        step_status             = 'NA'
        step_name               = 'NA'
        timestamp_action_start  = timestamp_utc_now()
        for action_step in self.action_steps:
            timestamp_start = timestamp_utc_now()
            step_name       = action_step.__name__
            try:
                step_data   = action_step()
                step_status = step_data.get('status')
            except Exception as e:
                step_status = 'exception'
                step_data   = {'exception': str(e) }

            timestamp_end = timestamp_utc_now()
            step_result = dict(name            = step_name           ,
                               step_data       = step_data           ,
                               step_status     = step_status         ,
                               timestamp_start = timestamp_start     ,
                               timestamp_end   = timestamp_end       )
            steps_result.append(step_result)
            if step_status != 'ok':
                break
        timestamp_action_end = timestamp_utc_now()

        if step_status == 'ok':
            action_status  = 'ok'
            action_message = f'action executed ok: {self.action_name}'
        elif step_status == 'exception':
            action_status = 'failed'
            action_message = step_data.get('exception')
        else:
            action_status  = 'failed'
            action_message = f'action failed :{self.action_name}'

        result = dict(action_id       = self.action_id         ,
                      action_name     = self.action_name       ,
                      action_message  = action_message         ,
                      action_outcome  = self.action_outcome    ,
                      action_status   = action_status          ,
                      last_step       = step_name              ,
                      timestamp_start = timestamp_action_start ,
                      timestamp_end   = timestamp_action_end   ,
                      steps_result    = steps_result           ,
                      task_id         = self.task_id           ,
                      workflow_id     = self.workflow_id       )
        return result

    def supported_model(self, model_provider, model):
        if model_provider == LLM__Providers.OPEN_ROUTER:
            if model in LLM__Models__Open_Router._value2member_map_:
                return True
        return False

    def step_1__check_llm_request(self):
        model          = self.llm_request.model
        model_provider = self.llm_request.model_provider
        user_prompt    = self.llm_request.user_prompt

        if not self.supported_model(model_provider, model):
            raise ValueError(f'Request provider {model_provider} and model {model} not supported')

        if not user_prompt:
            raise ValueError(f'no user prompt provided')

        if self.llm_chat_completion is None:
            raise ValueError(f'target llm_chat_completion was not set')

        if not isinstance(self.llm_chat_completion, LLM__Chat_Completion):
            raise ValueError(f'self.llm_chat_completion had the wrong type, it needs to implement LLM__Chat_Completion')

        step_data = dict(llm_request = self.llm_request.json())
        return status_ok(data=step_data)

    def step_2__create_llm_json_data(self):
        self.llm_chat_completion.add_messages__system(self.llm_request.system_prompts)
        self.llm_chat_completion.add_message__user   (self.llm_request.user_prompt   )
        json_data = self.llm_chat_completion.json_data()
        return status_ok(data=json_data)

    def step_3__llm_chat_completion(self):
        self.llm_response = self.llm_chat_completion.send_post_data()            # todo: add check for LLM response
        step_data = dict(llm_response = self.llm_response)
        return status_ok(data=step_data)

    def step_4__extract_content(self):
        if type(self.llm_response) is dict:
            choices = self.llm_response.get('choices')
            if choices and len(choices) > 0:
                choice  = choices[0]
                content = choice.get('message', {}).get('content')
                if content:                                         # todo add more checks here, to deal with edge cases
                    self.action_outcome = {'content': content}
                    return status_ok(data=self.action_outcome)
        return status_error(error='could not extract content from llm response')



from unittest                                        import TestCase
from osbot_aws.testing.Temp__Random__AWS_Credentials import Temp__Random__AWS_Credentials
from osbot_utils.utils.Env                           import get_env
from osbot_llms.backend.s3_minio.S3_DB_Base          import ENV_NAME__USE_MINIO_AS_S3, S3_DB_Base

class TestCase__S3_Minio__Temp_S3_Bucket(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.extra_env_vars   = { ENV_NAME__USE_MINIO_AS_S3: 'True'}
        cls.random_aws_creds  = Temp__Random__AWS_Credentials(env_vars=cls.extra_env_vars).set_vars()
        cls.s3_db_base        = S3_DB_Base(use_minio=True)
        assert get_env(ENV_NAME__USE_MINIO_AS_S3) == 'True'  # confirm value has been set
        with cls.s3_db_base as _:
            assert _.using_minio() is True                      # confirm we are using Minio
            _.setup()                                           # this will create the temp bucket
            assert _.bucket_exists() is True

    @classmethod
    def tearDownClass(cls):
        with cls.s3_db_base as _:
            assert _.bucket_delete_all_files() is True
            assert _.bucket_delete          () is True
        cls.random_aws_creds.restore_vars()



from osbot_fast_api.api.Fast_API            import Fast_API
from osbot_llms.fast_api.routes.Routes__Chat import Routes__Chat
from osbot_llms.fast_api.routes.Routes__Info import Routes__Info


class Fast_API__LLMs(Fast_API):
    base_path : str  = '/llms'

    def setup_routes(self):
        self.add_routes(Routes__Chat)
        self.add_routes(Routes__Info)



from osbot_utils.base_classes.Type_Safe     import Type_Safe
from osbot_llms.OSBot_LLMs__Shared_Objects  import osbot_llms__shared_objects


class S3__Chat_Threads(Type_Safe):

    def s3_db_chat_threads(self):
        return osbot_llms__shared_objects.s3_db_chat_threads()

    def all_chat_threads_for_today(self):
        return self.all_chat_threads_for_day()

    def all_chat_threads_for_day(self, day=None):
        with self.s3_db_chat_threads()  as _:
            s3_folder = _.s3_key_generator.s3_folder__for_day(day)
            return self.s3_db_chat_threads().s3_folder_files(s3_folder, include_sub_folders=True)



from os import getenv
from pathlib import Path

import openai
import requests
from openai                                         import AsyncOpenAI, OpenAI
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.testing.Temp_File                  import Temp_File
from osbot_utils.utils.Dev                          import pprint
from osbot_utils.utils.Env                          import load_dotenv
from osbot_utils.utils.Files                        import file_create_from_bytes


#URL_OPEN_AI_BASE   = "https://api.openai.com/v1"
OPEN_AI__API_KEY   = 'OPEN_AI__API_KEY'
DEFAULT_MAX_TOKENS = 4096
#MAX_ANSWER_SIZE = 512
#MAX_ANSWER_SIZE = 1024                              # todo: see impact of changing this from 512 to 1024
MAX_ANSWER_SIZE = 65536                              # this in practice removes any size restrictions # todo: add feature to enable this only for the billing mode


class API_Open_AI:

    def __init__(self):
        self.stream              = True
        self.seed                = 42
        self.temperature         = 0.0
        # self.model              = 'gpt-4-vision-preview'  # in Nov 2023 is there is a limit of 100 images per day
        self.model               =  'gpt-4o' #'gpt-4-1106-preview'  # gpt-3.5-turbo' #'gpt-4' #
        self.print_create_kwargs = False
        #self.logging             = Logging()

    @cache_on_self
    def api_key(self):
        load_dotenv()
        return getenv(OPEN_AI__API_KEY)

    def embeddings(self, input, model='text-embedding-3-small', dimensions=None):
        url       = 'https://api.openai.com/v1/embeddings'
        headers   = { "Content-Type" : "application/json"        ,
                      "Authorization": f"Bearer {self.api_key()}"}
        json_data = { "input"     : input,
                      "model"     : model }
        if dimensions:
            json_data["dimensions"] =  dimensions

        response        = requests.post(url=url, headers=headers, json=json_data)
        json_data       = response.json()
        embeddings_data = json_data.get('data')[0].get('embedding')
        total_tokens    = json_data.get('usage').get('total_tokens')
        result          = dict(embeddings_data = embeddings_data,
                               model           = model          ,
                               total_tokens     = total_tokens  )
        return result

    def open_ai_available(self):
        if self.api_key():
            #if is_url_online(URL_OPEN_AI_BASE):            # todo, find a better way (or url) to do this
                return True
        return False

    def open_ai_not_available(self):
        return self.open_ai_available() is False

    async def create(self, messages, model=None, temperature=None, seed=None, max_tokens=None):
        openai.api_key = self.api_key()
        kwargs = dict(model       = model       or self.model      ,
                      messages    = messages                       ,
                      temperature = temperature or self.temperature,
                      seed        = seed        or self.seed       ,
                      stream      = self.stream                    ,
                      max_tokens  = max_tokens  or DEFAULT_MAX_TOKENS)

        if self.print_create_kwargs:                            # todo : remove
           pprint(kwargs)

        #self.logging.add_open_ai_create_completions(kwargs)

        #client = AsyncOpenAI(api_key=self.api_key())                   # BUG: nasty bug that was causing loop not closed errors
        async with AsyncOpenAI(api_key=self.api_key()) as client:
            response = await client.chat.completions.create(**kwargs)

            async for chunk in response:
                content = chunk.choices[0].delta.content
                if content:
                    yield content
    def messages(self):
        return [{"role": "user", "content": 'Hi'}]


    def setup(self):
        openai.api_key = self.api_key()
        return self

    def ask_one_question_no_history(self,question, model=None, async_mode=False):
        messages    = [{"role": "user", "content": question}]
        return self.ask_using_messages(messages, model=model, async_mode=async_mode)

    async def ask_using_messages(self, messages, model=None, temperature=None, seed=None, max_tokens=None,  async_mode=False):
        generator    = self.create(messages, model=model, temperature=temperature,seed=seed, max_tokens=max_tokens)
        if async_mode:
            return generator
        full_answer = ""

        async for item in generator:
            if item is not None:
                full_answer += item
        return full_answer


    async def ask_using_system_prompts(self, user_prompt, images=None, system_prompts=None, histories=None, model=None, temperature=None, seed=None, max_tokens=None, async_mode=False):
        messages = []
        if system_prompts:
            for system_prompt in system_prompts:
                messages.append({"role": "system", "content": system_prompt})
        if histories:
            for item in histories:
                question = item.question
                answer   = item.answer
                if answer and len(answer) > MAX_ANSWER_SIZE:
                    answer = answer[:MAX_ANSWER_SIZE] + "...(max)..."
                messages.append({"role": "user"     , "content": question})
                messages.append({"role": "assistant", "content": answer})
        if images:
            user_prompt  = [ { "type"      : "text"      ,
                               "text"      : user_prompt }]
            for image_url in images:
                image = { "type"     : "image_url"       ,
                         "image_url": { "url": image_url }}
                user_prompt.append(image)
        messages.append({"role": "user", "content": user_prompt})

        return await self.ask_using_messages(messages, model=model, temperature=temperature, seed=seed, max_tokens=max_tokens,  async_mode=async_mode)

    def ask_question_with_user_data_and_prompt(self,user_question, user_data, system_prompt, user_history):
        messages = [{"role": "system", "content": system_prompt},
                    {"role": "system", "content": user_data    }]
        for item in user_history:
            question = item.get('question')
            answer   = item.get('answer')
            messages.append({"role": "user"     , "content": question})
            messages.append({"role": "assistant", "content": answer})
        messages.append({"role": "user", "content": user_question})

        #pprint(messages)
        return self.ask_using_messages(messages)


    def audio_to_text(self, audio_bytes, model="whisper-1"):
        with Temp_File(extension=".mp3") as temp_file:
            file_create_from_bytes(path=temp_file.file_path, bytes=audio_bytes)
            client        = OpenAI(api_key=self.api_key())  # refactor to helper method
            path          = Path(temp_file.file_path)
            transcription = client.audio.transcriptions.create(model=model,file=path)
            return transcription.text

    def text_to_audio(self, input_text, voice='shimmer', model="tts-1"):
        client = OpenAI(api_key=self.api_key())                                  # refactor to helper method
        response = client.audio.speech.create(model=model     ,                  # tts-1-hd | tts-1
                                              voice=voice     ,                  # alloy, echo, fable, onyx, nova, and shimmer
                                              input=input_text
        )
        audio_bytes = response.content
        return audio_bytes
        #response.stream_to_file(speech_file_path)